#include <iostream>
#include <iomanip>
#include <vector>
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/features2d/features2d.hpp>
#include <opencv2/calib3d/calib3d.hpp>


using namespace std;
using namespace cv;

// input points
vector<vector<Point3f>> objectPoints;  //vetor de opontos nas coordenadas dos objetos
vector<vector<Point2f>> imagePoints;   //vetor de pontos referentes na imagem
Mat cameraMatrix;   //parametros intrinsecos
Mat distCoeffs;     //coeficientes de distorção


int addChessboardPoints(vector<string>& filelist, Size & boardSize);
Mat  remap(Mat &image);
double calibrate(cv::Size imageSize);

int main()
{

    namedWindow("Image");
	Mat image;
	vector<string> filelist;
    Size boardSize(6,4);


	// gera a lista de arquivos a serem utilizados para detectar o tabuleiro
	for (int i=1; i<=20; i++) {
		stringstream str;
		str << "../chessboards/chessboard" << setw(2) << setfill('0') << i << ".jpg";
		cout << str.str() << endl;

		filelist.push_back(str.str());
		//image= imread(str.str(),0);
		//imshow("Image",image);
		//waitKey(100);
	}


    // adiciona pontos Object x Imagem
    addChessboardPoints(filelist,boardSize);
    
    
    // calibra a camera
    calibrate(Size(320,240));
    //poderia salvar a matriz intrinseca

    
    //apresenta a matriz de parâmetros intrinsecos
    std::cout << " Camera intrinsic: " << cameraMatrix.rows << "x" << cameraMatrix.cols << std::endl;
    std::cout << cameraMatrix.at<double>(0,0) << " " << cameraMatrix.at<double>(0,1) << " " << cameraMatrix.at<double>(0,2) << std::endl;
    std::cout << cameraMatrix.at<double>(1,0) << " " << cameraMatrix.at<double>(1,1) << " " << cameraMatrix.at<double>(1,2) << std::endl;
    std::cout << cameraMatrix.at<double>(2,0) << " " << cameraMatrix.at<double>(2,1) << " " << cameraMatrix.at<double>(2,2) << std::endl;
    
    
    // Image distorcida
    image = imread(filelist[10]);
	Mat uImage= remap(image);
    imshow("Original", image);
    imshow("Após correção de distorções", uImage);
    waitKey();
    
    
    //com a camera calibrada, vamos estimar para uma imagem a matriz de rotação e translação
    //depois disso, tentar desenhar nessa imagem
    Mat rvec;
    Mat tvec;
    
    //gera o modelo para uma imagem
    vector<Point3f> objectCorners;
    for (int i=0; i<boardSize.height; i++) {
        for (int j=0; j<boardSize.width; j++) {
            objectCorners.push_back(cv::Point3f(i, j, 0.0f));
        }
    }
    Mat objectPointsMat(objectCorners);
    
    
    for (int i=0; i<20; i++) {
        image = imread(filelist[i]);

        //acha os pontos correspondentes do modelo na imagem
        vector<Point2f> imageCorners;
        findChessboardCorners(image, boardSize, imageCorners);
        Mat imagePointsMat(imageCorners);
    
    
        solvePnP(objectPointsMat, imagePointsMat, cameraMatrix, distCoeffs, rvec, tvec);
    
        //plot
        vector<Point2f> pontos2d;

        //projeta os pontos 3D sobre a imagem
        projectPoints(objectCorners, rvec, tvec, cameraMatrix, distCoeffs, pontos2d);
        

        for (int i=0; i<objectCorners.size(); i++) {
            //desenha círuclos nos lugares
            circle(image,pontos2d[i], 3, Scalar(0, 255, 0));
        }
    
        imshow("Projected Image", image);
        waitKey();
    }
	return 0;
}


int addChessboardPoints(vector<string>& filelist, Size & boardSize) {
    
    vector<Point2f> imageCorners;
    vector<Point3f> objectCorners;
    
    // 3D Object Points:
    // localizações no tabuleiro (X,Y,Z)= (i,j,0)
    for (int i=0; i<boardSize.height; i++) {
        for (int j=0; j<boardSize.width; j++) {
            objectCorners.push_back(Point3f(i, j, 0.0f));
        }
    }
    
    // 2D Image points:
    Mat image; // to contain chessboard image
    int successes = 0;

    // para todas as imagens
    for (int i=0; i<filelist.size(); i++) {
        
        // Abre a imagem
        image = cv::imread(filelist[i],0);
        
        // Encontra os cantos do tabuleiro
        bool found = findChessboardCorners(image, boardSize, imageCorners);
        
        // obtem os cantos com números flutuantes
        cornerSubPix(image, imageCorners,
                         Size(5,5),
                         Size(-1,-1),
                         TermCriteria(TermCriteria::MAX_ITER +
                                          TermCriteria::EPS,
                                          30,		 // número máximo de interações
                                          0.1));     // acurácia
        
        // se existir uma detecção com a mesma quantidade de pontos do modelo
        if (imageCorners.size() == boardSize.area()) {
            
            // adiciona pontos dos objeto e da imagem
            imagePoints.push_back(imageCorners);
            objectPoints.push_back(objectCorners);
            successes++;
        }
        
        //Draw the corners
        //drawChessboardCorners(image, boardSize, imageCorners, found);
        //imshow("Corners on Chessboard", image);
        //waitKey(100);
    }
    
    return successes;
}


// calibra a camera
double calibrate(Size imageSize)
{
    vector<Mat> rvecs, tvecs;   //matries de rotação e translação (1 para cada imagem durante a calibração)

    return
    calibrateCamera(objectPoints, // pontos do objeto em 3D
                    imagePoints,  // pontos da imagem em 3D
                    imageSize,    // tamanho da imagem
                    cameraMatrix, // parametros intrinsecos
                    distCoeffs,   // parametros de distorção
                    rvecs, tvecs, // Rs, Ts
                    0);        // set options
    //			    ,CV_CALIB_USE_INTRINSIC_GUESS);
    
}

// remove distortion in an image (after calibration)
Mat  remap(Mat &image) {
    
    Mat undistorted;
    Mat map1,map2;
    
    
    initUndistortRectifyMap(
                                cameraMatrix,  // parâmetros intrinsecos
                                distCoeffs,    // parametros de distorção
                                cv::Mat(),     //
                                cv::Mat(),     //
                                cv::Size(640,480),
                                //            image.size(),  // size of undistorted
                                CV_32FC1,      // tipo de saida
                                map1, map2);   // the x and y mapping functions
    
    
    cv::remap(image, undistorted, map1, map2, 
              cv::INTER_LINEAR); // método de interpolação
    
    return undistorted;
}

